(* $Id: godi_match_lex.mll 748 2008-05-21 15:28:12Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

{
  exception Bracket_Unsupported
  exception Lexing_Error

  type bracket_token =
      Bracket_char of char
    | Bracket_range of (char * char)
    | Bracket_end

  type brace_token =
      Brace_literal of string
    | Brace_comma
    | Brace_braces of brace_token list  (* inner braces *)
    | Brace_end

  type glob_features =
      { enable_star : bool;
	enable_qmark : bool;
	enable_brackets : bool;
	enable_braces : bool;
	enable_escape : bool;
	mutable escaped : bool;
      }

  type glob_token =
      Glob_literal of string
    | Glob_star
    | Glob_qmark
    | Glob_brackets of (bool * bracket_token list)
    | Glob_braces of brace_token list
    | Glob_end

  type exploded_char =
      C of char   (* An unescaped character *)
    | E of char   (* An escaped character *)


  type regex_token =
      RE_alt        (* | *)
    | RE_many       (* * *)
    | RE_many1      (* + *)
    | RE_opt        (* ? *)
    | RE_bound of (int * int option)  (* { ints are: minimum, maximum *)
    | RE_bracket of (bool * bracket_token list) 
	            (* [ The bool says whether the expression is negated *)
    | RE_lparen     (* ( *)
    | RE_rparen     (* ) *)
    | RE_any        (* . *)
    | RE_lanchor    (* ^ *)
    | RE_ranchor    (* $ *)
    | RE_backref of char (* backslash + digit *)
    | RE_char of char  (* <character> *)
    | RE_end


  let rec collect_until end_token parse_fun lexbuf =
    let tok = parse_fun lexbuf in
    if tok = end_token then
      []
    else
      tok :: (collect_until end_token parse_fun lexbuf)

}

(* bracket_rest: Scans a bracket expression beginning at the second 
 * character (where ']' is always the terminating character)
 *)

rule bracket_rest = parse
    "[:" [^ ':' ] ":]" { raise Bracket_Unsupported }
  | "[." [^ '.' ] ".]" { raise Bracket_Unsupported }
  | "[=" [^ '=' ] "=]" { raise Bracket_Unsupported }
  | "]"                { Bracket_end }
  | [ ^ ']' ] "-" [^ ']' ]     
                       { let c0 = Lexing.lexeme_char lexbuf 0 in
			 let c1 = Lexing.lexeme_char lexbuf 2 in
			 if c0 > c1 then raise Lexing_Error;
			 Bracket_range(c0,c1)
		       }
  | eof                { raise Lexing_Error }
  | [ ^ ']' ]          { Bracket_char (Lexing.lexeme_char lexbuf 0) }

(* bracket_first: Scans the first token of a bracket expression
 * (after "[", "[^", or "[!").
 * Here, ']' is not recognized as terminating character.
 *)

and bracket_first = parse
    "[:" [^ ':' ] ":]" { raise Bracket_Unsupported }
  | "[." [^ '.' ] ".]" { raise Bracket_Unsupported }
  | "[=" [^ '=' ] "=]" { raise Bracket_Unsupported }
  | _ "-" [^ ']' ]     { let c0 = Lexing.lexeme_char lexbuf 0 in
			 let c1 = Lexing.lexeme_char lexbuf 2 in
			 if c0 > c1 then raise Lexing_Error;
			 Bracket_range(c0,c1)
		       }
  | eof                { raise Lexing_Error }
  | _                  { Bracket_char (Lexing.lexeme_char lexbuf 0) }


(* brace: Collects material within brace expressions (case: backslash
 * is escape character
 *)

and brace = parse
    "}"                { Brace_end }
  | ","                { Brace_comma }
  | "{"                { let l = collect_until Brace_end brace lexbuf in
			 Brace_braces l }
  | '\\' _             { Brace_literal (Lexing.lexeme lexbuf) }
  | [^ '}' ',' '\\' '{' ]  { Brace_literal (Lexing.lexeme lexbuf) }
  | eof                { raise Lexing_Error }
  | _                  { raise Lexing_Error }

(* brace_noescape: Used for the case that backslash is not an escape
 * character
 *)

and brace_noescape = parse
    "}"                { Brace_end }
  | ","                { Brace_comma }
  | "{"                { let l = collect_until Brace_end brace_noescape lexbuf in
			 Brace_braces l }
  | [^ '}' ',' '{']    { Brace_literal (Lexing.lexeme lexbuf) }
  | eof                { raise Lexing_Error }
  | _                  { raise Lexing_Error }

and glob_expr feat = parse
    "*"                { if feat.enable_star && not feat.escaped then 
			   Glob_star
			 else (
			   feat.escaped <- false;
			   Glob_literal "*"
			 )
		       }
  | "?"                { if feat.enable_qmark && not feat.escaped then 
			   Glob_qmark
			 else (
			   feat.escaped <- false;
			   Glob_literal "?"
			 )
		       }
  | "[" [ '!' '^' ]?   { if feat.enable_brackets && not feat.escaped then (
			   let negated = 
			     String.length(Lexing.lexeme lexbuf) > 1 in
			   let t0 = bracket_first lexbuf in
			   let l = collect_until 
				     Bracket_end bracket_rest lexbuf in
			   Glob_brackets (negated, t0 :: l)
			 )
			 else (
			   feat.escaped <- false;
			   Glob_literal (Lexing.lexeme lexbuf)
			 )
		       }
  | "{"                { if feat.enable_braces && not feat.escaped then (
			   let p =
			     if feat.enable_escape then
			       brace
			     else
			       brace_noescape in
			   let l = collect_until Brace_end p lexbuf in
			   Glob_braces l
			 )
			 else (
			   feat.escaped <- false;
			   Glob_literal "{"
			 )
		       }
  | "\\"               { if feat.enable_escape && not feat.escaped then (
			   feat.escaped <- true;
			   Glob_literal ""
			 )
			 else (
			   feat.escaped <- false;
			   Glob_literal "\\"
			 )
		       }
  | [ ^ '*' '?' '[' '{' '\\' ]+ 
                       { feat.escaped <- false;
			 Glob_literal (Lexing.lexeme lexbuf)
		       }
  | eof                { if feat.escaped then raise Lexing_Error;
			 Glob_end
		       }

and generic_lex = parse
    '\\' _             { let char = E (Lexing.lexeme_char lexbuf 1) in
                         char :: generic_lex lexbuf }
  | _                  { let char = C (Lexing.lexeme_char lexbuf 0) in
                         char :: generic_lex lexbuf }
  | eof                { [] }

and generic_lex_noescape = parse
  | _                  { let char = C (Lexing.lexeme_char lexbuf 0) in
                         char :: generic_lex_noescape lexbuf }
  | eof                { [] }


and ext_regex_expr = parse
    '|' { RE_alt }
  | '*' { RE_many }
  | '+' { RE_many1 }
  | '?' { RE_opt }
  | '{' [ '0'-'9' ]+ ( ',' ( [ '0'-'9' ]+ )? )? '}'
    { let s = Lexing.lexeme lexbuf in
      try
	let comma = String.index s ',' in
	let n_min = int_of_string(String.sub s 1 (comma - 1)) in
	let n_max =
	  if s.[comma+1] = '}' then
	    None
	  else
	    Some(int_of_string(String.sub s (comma+1) (String.length s - comma - 2))) in
	RE_bound(n_min,n_max)
      with
	  Not_found ->
	    let n = int_of_string(String.sub s 1 (String.length s - 2)) in
	    RE_bound(n,Some n)
    }
  | '[' '^'?   (* plus the whole bracket expr until ']' *)
    {
      let negated = 
	String.length(Lexing.lexeme lexbuf) > 1 in
      let t0 = bracket_first lexbuf in
      let l = collect_until 
		Bracket_end bracket_rest lexbuf in
      RE_bracket (negated, t0 :: l)
    } 
  | '(' { RE_lparen }
  | ')' { RE_rparen }
  | '.' { RE_any }
  | '^' { RE_lanchor }
  | '$' { RE_ranchor }
  | '\\' [ ^ 'A'-'Z' 'a'-'z' ]
    { RE_char (Lexing.lexeme_char lexbuf 1) }
  | '\\' _
    { raise Lexing_Error }
  | '\\' eof
    { raise Lexing_Error }
  | _ 
    { RE_char (Lexing.lexeme_char lexbuf 0) }
  | eof
    { RE_end }

and basic_regex_expr = parse
  | '*' { RE_many }
  | "\\{" [ '0'-'9' ]+ ( ',' ( [ '0'-'9' ]+ )? )? "\\}"
    { let s = Lexing.lexeme lexbuf in
      try
	let comma = String.index s ',' in
	let n_min = int_of_string(String.sub s 2 (comma - 2)) in
	let n_max =
	  if s.[comma+1] = '\\' then
	    None
	  else
	    Some(int_of_string(String.sub s (comma+1) (String.length s - comma - 3))) in
	RE_bound(n_min,n_max)
      with
	  Not_found ->
	    let n = int_of_string(String.sub s 2 (String.length s - 4)) in
	    RE_bound(n,Some n)
    }
  | '[' '^'?   (* plus the whole bracket expr until ']' *)
    {
      let negated = 
	String.length(Lexing.lexeme lexbuf) > 1 in
      let t0 = bracket_first lexbuf in
      let l = collect_until 
		Bracket_end bracket_rest lexbuf in
      RE_bracket (negated, t0 :: l)
    } 
  | "\\(" { RE_lparen }
  | "\\)" { RE_rparen }
  | '.' { RE_any }
  | '^' { RE_lanchor }
  | '$' { RE_ranchor }
  | '\\' [ ^ 'A'-'Z' 'a'-'z' '1'-'9' ]
    { RE_char (Lexing.lexeme_char lexbuf 1) }
  | '\\' [ '1'-'9' ] 
    { RE_backref (Lexing.lexeme_char lexbuf 1) }   
  | '\\' _
    { raise Lexing_Error }
  | '\\' eof
    { raise Lexing_Error }
  | _ 
    { RE_char (Lexing.lexeme_char lexbuf 0) }
  | eof
    { RE_end }


and parse_shell_words = parse
  | [ ' ' '\t' '\r' '\n' ]+  { parse_shell_words lexbuf }
  | '"'                      { let (word,words) = parse_shell_dquote lexbuf in
			       word :: words
			     }
  | '\''                     { let (word,words) = parse_shell_squote lexbuf in
			       word :: words
			     }
  | '\\' '\n'                { parse_shell_words lexbuf }
  | '\\' ( _ as c)           { let (word,words) = parse_shell_word lexbuf in
			       (String.make 1 c ^ word) :: words
			     }
  | [ ^ '\\'] as c           { let (word,words) = parse_shell_word lexbuf in
			       (String.make 1 c ^ word) :: words
			     }
  | eof                      { [] }

and parse_shell_word = parse
  | [ ' ' '\t' '\r' '\n' ]+  { ("", parse_shell_words lexbuf) }
  | '"'                      { parse_shell_dquote lexbuf }
  | '\''                     { parse_shell_squote lexbuf }
  | '\\' '\n'                { parse_shell_word lexbuf }
  | '\\' ( _ as c)           { let (word,words) = parse_shell_word lexbuf in
			       ((String.make 1 c ^ word), words)
			     }
  | [ ^ '\\' '"' '\'' ' ' '\t' '\r' '\n' ]+ as c 
                             { let (word,words) = parse_shell_word lexbuf in
			       ((c ^ word), words)
			     }
  | eof                      { ("", []) }

and parse_shell_dquote = parse
  | '"'                      { ("", parse_shell_words lexbuf ) }
  | '\\' '\n'                { parse_shell_dquote lexbuf }
  | '\\' (_ as c)            { let (word,words) = parse_shell_dquote lexbuf in
			       ((String.make 1 c ^ word), words)
			     }
  | [ ^ '\\' '"']+ as c      { let (word,words) = parse_shell_dquote lexbuf in
			       ((c ^ word), words)
			     }

and parse_shell_squote = parse
  | '\''                     { ("", parse_shell_words lexbuf ) }
  | [^ '\'']+ as c           { let (word,words) = parse_shell_squote lexbuf in
			       ((c ^ word), words)
			     }
