(* $Id: godi_script.ml 578 2005-08-10 11:08:38Z gerd $ *)

open Printf

let runcmd s = 
  flush stdout;
  flush stderr;
  Sys.command s ;;

let runcmdf fmt =
  kprintf runcmd fmt ;;

let read_from_fd f = 
  let b = Buffer.create 200 in
  let s = String.create 4096 in
  try
    while true do
      let n = input f s 0 (String.length s) in
      if n=0 then raise End_of_file;
      Buffer.add_substring b s 0 n;
    done;
    assert false
  with
      End_of_file ->
	Buffer.contents b
;;


let read_file name =
  let f = open_in name in
  try
    let s = read_from_fd f in
    close_in f;
    s
  with
      error -> close_in f; raise error
;;


let read_from_cmd cmdstr =
  let f = Unix.open_process_in cmdstr in
  let s =
    try
      read_from_fd f
    with
	error -> ignore(Unix.close_process_in f); raise error in
  let p = Unix.close_process_in f in
  match p with
      Unix.WEXITED n -> 
	(n, s)
    | Unix.WSIGNALED n ->
	failwith("Command was killed by signal " ^ string_of_int n)
    | Unix.WSTOPPED n ->
	(* We normally do not get this case *)
	assert false
;;


let read_from_cmdf fmt = 
  kprintf read_from_cmd fmt ;;


let write_file name str =
  let fd = open_out name in
  try
    output_string fd str;
    close_out fd
  with
      error -> close_out fd; raise error
;;

	
let env name =
  try Sys.getenv name with Not_found -> "" ;;

let env_opt name =
  let e = env name in
  if e = "" then None else Some e ;;

let env_req name =
  match env_opt name with
      Some e -> e
    | None -> failwith ("Required environment variable " ^ name ^ " is not set")
;;


let ws_re = Str.regexp "[ \t\r\n]+";;
let colon_re = Str.regexp ":" ;;

let findcmd ?(path = Str.split colon_re (env_req "PATH")) cmdstr =
  match
    List.fold_left
      (fun acc dir ->
	 match acc with
	     None ->
	       ( try 
		   let f = Filename.concat dir cmdstr in
		   let s = Unix.stat f in
		   if s.Unix.st_kind <> Unix.S_REG then raise Not_found;
		   Unix.access f [Unix.X_OK];
		   Some f
		 with
		     Not_found
		   | Unix.Unix_error(_,_,_) -> acc
	       )
	   | Some _ -> acc
      )
      None
      path
  with
      Some f -> f
    | None -> raise Not_found
;;


let system_path() =
  Str.split colon_re (env_req "PATH") ;;

let search_libs_path() =
  let p = Str.split ws_re (env_req "SEARCH_LIBS") in
  List.map (fun d -> Filename.concat d "bin") p
;;


type 'a cmdstate =
    { code : int;
      result : 'a option
    }

let code st = st.code ;;
let result st = st.result ;;

let return_code n st = { st with code = n };;
let return_result x st = { st with result = x };;

let initial_state () = { code = 0; result = None };;


type 'a cmdlet = 'a cmdstate -> 'a cmdstate ;;

let cmd cmdstr st =
  let n = runcmd cmdstr in
  { st with code = n }
;;

let cmdf fmt = kprintf cmd fmt ;;


let cmd_output cmdstr st =
  let (n, s) = read_from_cmd cmdstr in
  { st with code = n; result = Some s }
;;

let cmdf_output fmt = kprintf cmd_output fmt ;;

let set_code n st =
  { st with code = n }
;;

let set_code_from f st =
  { st with code = f st }
;;

let set_bool_code b st =
  { st with code = if b then 0 else 1 }
;;

let set_bool_code_from f st =
  { st with code = if f st then 0 else 1 }
;;

let ( &- ) cl1 cl2 st =
  let st1 = cl1 st in
  if st1.code = 0 then
    cl2 st1
  else
    st1
;;

let ( |- ) cl1 cl2 st =
  let st1 = cl1 st in
  if st1.code = 0 then
    st1
  else
    cl2 st1
;;

let ignore_code cl st =
  let st' = cl st in
  { st' with code = 0 }
;;

let eval cl =
  let st = initial_state() in
  let st' = cl st in
  match st' with
      { code = 0; result = Some r } -> Some r
    | _ -> None
;;

let eval_test cl =
  let st = initial_state() in
  let st' = cl st in
  st'.code = 0
;;

type lang_env =
    [ `C of c_env ]

and c_env = 
  { c_env_file : (string * string) option;
    c_incdirs : string list;
    c_libdirs : string list;
    c_libs : string list;
    c_flags : string list;
    c_ld_flags : string list;
    c_elf_rpath : bool;
    c_config_script : string option;
    c_godi_deps : string list;
    c_requirements : c_env list;
  }

let empty_c_env =
  { c_env_file = None;
    c_incdirs = [];
    c_libdirs = [];
    c_libs = [];
    c_flags = [];
    c_ld_flags = [];
    c_elf_rpath = false;
    c_config_script = None;
    c_godi_deps = [];
    c_requirements = [];
  } ;;


let create_test_whether_c_function_exists f_name =
  sprintf
    "char %s (); int main() { void * h; h = %s; return 0; }\n"
    f_name 
    f_name
;;


let temp_file prefix suffix =
  let rec create n =
    try
      n, (Unix.openfile 
	    (prefix ^ string_of_int n ^ suffix) [ Unix.O_WRONLY;
						  Unix.O_CREAT;
						  Unix.O_EXCL ] 0o666)
    with
	Unix.Unix_error(Unix.EEXIST,_,_) ->
	  create(n+1)
  in
  let (n, fd) = create 0 in
  Unix.close fd;
  prefix ^ string_of_int n ^ suffix
;;


let log_started = ref false ;;

let log msg =
  let m = if !log_started then Open_append else Open_trunc in
  let fd = open_out_gen
	     [ Open_wronly; Open_creat; m ] 0o666 "script.log" in
  log_started := true;
  try
    output_string fd msg;
    close_out fd;
  with
      error -> close_out fd; raise error
;;


let logf format =
  kprintf log format ;;

let dollar_re = Str.regexp "[$]";;

let print_conf_file file vars =
  (* Known bugs: We cannot write values with spaces at the beginning or at
   * the end. Also, a backslash at the end will not work.
   *)
  let escape_dollar s =
    Str.global_replace dollar_re "$$" s in
  let s =
    String.concat 
      "\n" 
      (List.map
	 (fun (name,value) ->
	    escape_dollar name ^ "=" ^ escape_dollar value
	 )
	 vars
      )
  in
  write_file file (s ^ "\n")
;;


let line_re = Str.regexp "\r?\n" ;;
let name_value_re = Str.regexp "^\\([^= \t]+\\)[ \t]*=[ \t]*\\(.*\\)$";;
let find_dollar_re = Str.regexp "[$]\\(.|$\\)";;
let empty_line_re = Str.regexp "^[#]\\|^[ \t]+$" ;;

let parse_conf_file file =
  let unescape_dollar s =
    Str.global_substitute 
      find_dollar_re 
      (fun _ ->
	 let k = Str.match_beginning() + 1 in
	 if k = String.length s || s.[k] <> '$' then
	   failwith (sprintf "File %s: Unsupported '$' reference" file);
	 "$"
      )
      s in

  let s = read_file file in
  let lines = Str.split line_re s in
  List.flatten
    (List.map
       (fun line ->
	  if line <> "" && line.[ String.length line - 1 ] = '\\' then
	    failwith (sprintf "File %s: cannot parse continuation lines" file);
	  if Str.string_match name_value_re line 0 then (
	    let name = Str.matched_group 1 line in
	    let value = Str.matched_group 2 line in
	    [ unescape_dollar name, unescape_dollar value ]
	  )
	  else
	    if Str.string_match empty_line_re line 0 then
	      []
	    else
	      failwith (sprintf "File %s: unknown syntax: %s" file line)
       )
       lines
    )
;;


let rec c_flags ?(with_requirements = false) ce =
  (* /usr/include is usually automatically found, so we can
   * delete it.
   *)
  let incdirs =
    List.filter (fun d -> d <> "/usr/include") ce.c_incdirs in

  let flags =
    List.map (fun d -> "-I" ^ d) incdirs @ 
    ce.c_flags in

  if with_requirements then
    let rflags = 
      List.flatten (List.map (c_flags ~with_requirements) ce.c_requirements)
    in
    rflags @ flags
  else
    flags
;;


let rec c_ld_flags ?(with_requirements = false) ce =
  (* /usr/lib is usually automatically found, so we can
   * delete it.
   *)
  let libdirs =
    List.filter (fun d -> d <> "/usr/lib") ce.c_libdirs in

  let flags =
    List.map (fun d -> "-L" ^ d) libdirs @ 
    ( if ce.c_elf_rpath then
	let f = env "ELF_RPATH_FLAG" in
	(List.map (fun d -> "-Wl," ^ f ^ d) libdirs)
      else
	[]) @
    List.map (fun lib -> "-l" ^ lib) ce.c_libs @
    ce.c_ld_flags in

  if with_requirements then
    let rflags = 
      List.flatten (List.map (c_ld_flags ~with_requirements) ce.c_requirements)
    in
    flags @ rflags
  else
    flags
;;


let c_compiler_name() = env_req "CC" ;;

let c_compile_test ce teststr st = 
  let early_flags =
    c_flags ~with_requirements:true ce in
  let late_flags =
    c_ld_flags ~with_requirements:true ce in
  let early_flag_string = 
    (* env "CFLAGS" ^ " " ^ *)   (* unclear whether we need it *)
    String.concat " " early_flags in
  let late_flag_string = 
    String.concat " " late_flags 
    (* ^ " " ^ env "LDFLAGS" *)  (* unclear whether we need it *)
  in
  let c_name = temp_file "test" ".c" in
  let base_name = Filename.chop_suffix c_name ".c" in
  write_file c_name teststr;
  let cc = env_req "CC" in
  let cmdstr =
    sprintf "%s -o %s %s %s.c %s 2>>script.log >&2"
      cc
      base_name
      early_flag_string
      base_name
      late_flag_string in
  logf "* Compiling:\n";
  logf "  %s\n" cmdstr;
  cmd cmdstr { st with result = Some base_name }
;;


let run_test st =
  match st.result with
      None -> 
	{ st with code = 2 }
    | Some file_name -> 
	let cmdstr = 
	  (Filename.concat Filename.current_dir_name file_name) ^
	  " 2>>script.log" in
	logf "* Running test:\n";
	logf " %s\n" cmdstr;
	cmd_output cmdstr st
;;


let find_c_library 
      ?(godi_deps = [])
      ?(pref_incdir = "")
      ?(pref_libdir = "")
      ?(flags = [])
      ?(ld_flags = [])
      ?inc_name 
      ?(requirements = [])
      ?(locations = Str.split (Str.regexp "[ \t]+") (env "SEARCH_LIBS"))
      ?(frameworks = [])
      ~libs
      ~test () =

  let print_test_case incdir libdir =
    match inc_name with
	None ->
	  let l = String.concat "+" libs in
	  printf "  Library=%s/<%s>\n" libdir l;
	  flush stdout
      | Some name ->
	  let l = String.concat "+" libs in
	  printf "  Include=%s/%s Library=%s/<%s>\n" incdir name libdir l;
	  flush stdout
  in

  let test_include incdir st =
    match inc_name with
	None ->
	  { st with code = 0}
      | Some name ->
	  let code = 
	    if Sys.file_exists (Filename.concat incdir name) then 0 else 1 in
	  { st with code = code }
  in

  let test_program_1 incdir libdir deps elf_rpath st =
    let ce =
      { c_env_file = None;
	c_incdirs = [ incdir ];
	c_libdirs = [ libdir ];
	c_libs = libs;
	c_flags = flags;
	c_ld_flags = ld_flags;
	c_elf_rpath = elf_rpath;
	c_config_script = None;
	c_godi_deps = deps;
	c_requirements = requirements;
      } in
    match test ce with
	None ->
	  { st with code = 1 }
      | Some ce' ->
	  { st with code = 0; result = Some ce' }
  in

  let test_program incdir libdir deps =
    test_program_1 incdir libdir deps false |- 
    test_program_1 incdir libdir deps true
  in 

  let have_preferences =
    set_bool_code (pref_incdir <> "" || pref_libdir <> "") in

  let check_preferences =
    have_preferences &-
    (fun st ->
       printf "Checking the suggestion\n";
       print_test_case pref_incdir pref_libdir; st) &-
    test_include pref_incdir &- 
    test_program pref_incdir pref_libdir godi_deps in

  let have_frameworks =
    set_bool_code (
      try
	frameworks <> [] &&
	Sys.getenv "SHLIB_TYPE" = "MachO"
      with
	| Not_found -> false) in
  (* We simply assume that MachO implies the existence of frameworks. Since
   * OS X is the only framework-based and the only MachO system, this test
   * works. In the future, this might change.
   *)

  let test_program_fw fwlist deps st =
    let ce =
      { c_env_file = None;
	c_incdirs = [ ];
	c_libdirs = [ ];
	c_libs = [ ];
	c_flags = flags;
	c_ld_flags = 
	  (List.map (fun fw -> "-framework " ^ fw) fwlist) @ ld_flags;
	c_elf_rpath = false;
	c_config_script = None;
	c_godi_deps = deps;
	c_requirements = requirements;
      } in
    match test ce with
	None ->
	  { st with code = 1 }
      | Some ce' ->
	  { st with code = 0; result = Some ce' }
  in

  let check_frameworks =
    have_frameworks &-
    (fun st ->
       printf "Checking frameworks\n";
       printf "  Framworks=%s\n" (String.concat "," frameworks);
       flush stdout;
       st) &-
    test_program_fw frameworks godi_deps in
  
  List.fold_left
    (fun acc dir ->
       let incdir = Filename.concat dir "include" in
       let libdir = Filename.concat dir "lib" in
       acc |- ( set_bool_code true &-
                (fun st ->
		   printf "Checking %s:\n" dir;
		   print_test_case incdir libdir; st) &-
		test_include incdir &-
		test_program incdir libdir [] )
    )
    (check_preferences |- check_frameworks)
    locations
;;


let print_c_result ?(vars = []) ~file ~prefix ce =
  printf "Directories for include files: %s\n" 
    (String.concat " " ce.c_incdirs);
  printf "Directories for libraries:     %s\n"
    (String.concat " " ce.c_libdirs);
  printf "Libraries:                     %s\n"
    (String.concat " " ce.c_libs);
  printf "Further CFLAGSs:               %s\n"
    (String.concat " " ce.c_flags);
  printf "Further LDFLAGS:               %s\n"
    (String.concat " " ce.c_ld_flags);
  printf "Needs RPATH:                   %s\n"
    (if ce.c_elf_rpath then "yes" else "no");
  printf "Used configure script:         %s\n"
    (match ce.c_config_script with Some s -> s | None -> "none");
  printf "Additional deps:               %s\n"
    (String.concat " " ce.c_godi_deps);
  flush stdout;

  let prep_reqs l =
    let l' = List.filter (fun c -> c.c_env_file <> None) l in
    List.map
      (fun c ->
	 match c.c_env_file with
	   | Some (abs_file_name, var_prefix) ->
	       var_prefix ^ ":" ^ abs_file_name
	   | None ->
	       assert false
      )
      l'
  in

  let incdirs =
    List.filter (fun d -> d <> "/usr/include") ce.c_incdirs in
  let libdirs =
    List.filter (fun d -> d <> "/usr/lib") ce.c_libdirs in

  let total_cflags = c_flags ce  in
  let total_ldflags = c_ld_flags ce in

  let vars' =
    [ prefix ^ "_INCDIRS", String.concat " " incdirs;
      prefix ^ "_LIBDIRS", String.concat " " libdirs;
      prefix ^ "_LIBNAMES", String.concat " " ce.c_libs;
      prefix ^ "_CFLAGS",  String.concat " " ce.c_flags;
      prefix ^ "_LDFLAGS", String.concat " " ce.c_ld_flags;
      prefix ^ "_NEED_RPATH", if ce.c_elf_rpath then "yes" else "no";
      prefix ^ "_TOTAL_CFLAGS",  String.concat " " total_cflags;
      prefix ^ "_TOTAL_LDFLAGS", String.concat " " total_ldflags;
      prefix ^ "_SCRIPT", 
                   (match ce.c_config_script with Some s -> s | None -> "");
      prefix ^ "_DEPENDS", String.concat " " ce.c_godi_deps;
      prefix ^ "_REQS", String.concat " " (prep_reqs ce.c_requirements);
    ] in
  print_conf_file file (vars' @ vars)
;;


let req_re = Str.regexp "\\([^:]+\\):\\(.*\\)$"


let rec parse_c_result ~file ~prefix () =
  let prefix_length = String.length prefix in
  let vars = parse_conf_file file in
  (* Strip prefix: *)
  let vars' =
    List.flatten
      (List.map
	 (fun (name, value) ->
	    if String.length name >= prefix_length &&
	       (String.sub name 0 prefix_length = prefix) then
		 [ String.sub 
		     name prefix_length (String.length name - prefix_length),
		   value ]
	    else
	      []
	 )
	 vars
      )
  in
  let extract name =
    try Str.split ws_re (List.assoc name vars') with Not_found -> [] in

  let reqs_l = extract "_REQS" in
  let reqs =
    List.map
      (fun req_spec ->
	 if Str.string_match req_re req_spec 0 then
	   let req_prefix = Str.matched_group 1 req_spec in
	   let req_file = Str.matched_group 2 req_spec in
	   parse_c_result ~file:req_file ~prefix:req_prefix ()
	 else
	   failwith "Cannot parse requirements"
      )
      reqs_l in

  { empty_c_env with
      c_env_file = Some (file, prefix);
      c_incdirs = extract "_INCDIRS" @ extract "_INCDIR";
      c_libdirs = extract "_LIBDIRS" @ extract "_LIBDIR";
      c_libs = extract "_LIBNAMES";
      c_flags = extract "_CFLAGS";
      c_ld_flags = extract "_LDFLAGS";
      c_elf_rpath = extract "_NEED_RPATH" = [ "yes" ];
      c_config_script =
        ( match extract "_SCRIPT" with [s] -> Some s | _ -> None );
      c_godi_deps = extract "_DEPENDS";
      c_requirements = reqs;
  }
;;
  


let main_c_finder ?godi_deps ?pref_incdir ?pref_libdir ?c_flags ?ld_flags 
                  () =
  let get_opt_arg which cl_name cl_help =
    match which with
	Some v -> 
	  [ cl_name, 
	    Arg.String (fun s -> if s <> "" then v := Some s), 
	    cl_help ]
      | None -> [] in

  let split = Str.split (Str.regexp "[ \t]+") in

  let get_list_arg which cl_name cl_help =
    match which with
	Some v -> [ cl_name, Arg.String (fun s -> v := split s), cl_help ]
      | None -> [] in

  Arg.parse
      ( get_list_arg 
	  godi_deps "-godi-deps" "<deps> Request additional dependencies" @
	
	get_opt_arg
	  pref_incdir "-pref-incdir" "<dir> Set preferred include directory" @

	get_opt_arg
	  pref_libdir "-pref-libdir" "<dir> Set preferred library directory" @

	get_list_arg
	  c_flags "-c-flags" "<flags> Set additional C compiler flags" @

	get_list_arg
	  ld_flags "-ld-flags" "<flags> Set additional linker flags"
      )
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "options:"
;;

	  
let c_env_from_config_script 
      ?(arg_cflags = "--cflags") ?(arg_ldflags = "--libs") ?(args = [])
      ~script_name () =

  let arg_str =
    String.concat " " (List.map Filename.quote args) in

  let cmd_cflags =
    sprintf "%s %s %s" script_name arg_cflags arg_str in

  logf "* Calling: %s\n" cmd_cflags;

  let (code, cflags) = read_from_cmd cmd_cflags in
  if code <> 0 then 
    failwith ("Error when running: " ^ cmd_cflags);

  let cmd_ldflags =
    sprintf "%s %s %s" script_name arg_ldflags arg_str in

  logf "* Calling: %s\n" cmd_ldflags;

  let (code, ldflags) = read_from_cmd cmd_ldflags in
  if code <> 0 then 
    failwith ("Error when running: " ^ cmd_ldflags);

  let split s = 
    Str.split ws_re s in

  { empty_c_env with
      c_flags = split cflags;
      c_ld_flags = split ldflags;
      c_config_script = Some script_name
  }
;;


let main_config_finder ~script_basename () =
  let pref_config = ref None in
 
  Arg.parse
      [ "-pref-config", 
	Arg.String (fun s -> if s <> "" then pref_config := Some s),
	"<path> Use this config script" ]
      (fun _ -> raise(Arg.Bad "Unexpected argument"))
      "options:";

  let script_name =
    match !pref_config with
	Some p -> 
	  if p = "" || p.[0] <> '/' then
	    failwith("This name is not absolute: " ^ p);
	  p
      | None ->
	  try
	    let path = search_libs_path() @ system_path() in
	    findcmd ~path script_basename
	  with
	      Not_found ->
		failwith ("Script " ^ script_basename ^ " not found")
  in

  script_name
;;
 
